#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sysexits.h>
#include <errno.h>

#include <sys/stat.h>
#include <sys/types.h>

#include <linux/input.h>
#include <linux/uinput.h>

#include <code-name.h>

/***************************************************************************/

#define JM_NAME_SIZE 128

#define JM_SLOW_STALL 190000
#define JM_FAST_STALL  50000

/* can't go above 999999 for slow, and really fast can be awkward too */
#define JM_SLOW_LIMIT 800000
#define JM_FAST_LIMIT   2000

#define JM_DEFAULT_UINPUT  "/dev/input/uinput"
#define JM_FALLBACK_UINPUT "/dev/misc/uinput"

/***************************************************************************/

static struct code_name button_table[] = {
 { "joystick", BTN_JOYSTICK },
 /* careful - joystick is the same as trigger */
 { "trigger",  BTN_TRIGGER },
 { "thumb",    BTN_THUMB },
 { "thumb2",   BTN_THUMB2 },
 { "top",      BTN_TOP },
 { "top2",     BTN_TOP2 },
 { "pinkie",   BTN_PINKIE },
 { "base",     BTN_BASE },
 { "base2",    BTN_BASE2 },
 { "base3",    BTN_BASE3 },
 { "base4",    BTN_BASE4 },
 { "base5",    BTN_BASE5 },
 { "base6",    BTN_BASE6 },
 {  NULL,      0 }
};

static struct code_name abs_table[] = {
  { "x",        ABS_X }, 
  { "y",        ABS_Y }, 
  /* these and others are completely untested */
  { "z",        ABS_Z }, 
  { "throttle", ABS_THROTTLE },
  { "rudder",   ABS_RUDDER },
  { "wheel",    ABS_WHEEL },
  { "gas",      ABS_GAS },
  { "hat0x",    ABS_HAT0X },
  { "hat0y",    ABS_HAT0Y },
  { "hat1x",    ABS_HAT1X },
  { "hat1y",    ABS_HAT1Y },
  { "hat2x",    ABS_HAT2X },
  { "hat2y",    ABS_HAT2Y },
  { "hat3x",    ABS_HAT3X },
  { "hat3y",    ABS_HAT3Y },
  {  NULL,     -1 }
};

static struct code_name cap_table[] = {
  { "syn",  EV_SYN },
  { "key",  EV_KEY },
  { "rel",  EV_REL },
  { "abs",  EV_ABS },
  { "msc",  EV_MSC },
  { "led",  EV_LED },
  { "snd",  EV_SND },
  { "rep",  EV_REP },
  { "ffd",  EV_FF  },
#ifdef EV_PWR
  { "pwr",  EV_PWR },
#endif
#ifdef EV_FF_STATUS
  { "ffs",  EV_FF_STATUS },
#endif
  { NULL,   -1 }
};

/***************************************************************************/

static void usage(char *app);
static int make_rodent(char *app, int make);
static int check_stick(char *app, int fd, int verbose);

/***************************************************************************/

static void usage(char *app)
{
  int i;

  printf("%s v%s: evdev+uinput joystick to mouse converter\n", app, VERSION);

  printf("\n");

  printf("usage: %s [-c] [-v] [-q] [-x] [-h] [-l key] [-r key] [-m key] [-b key] [-X axis] [-Y axis] /dev/input/event?\n", app);
  printf(" -c       copyright notice\n");
  printf(" -v       increase verbosity, can be used repeatedly\n");
  printf(" -q       run quietly\n");
  printf(" -x       ignore missing capabilities\n");
  printf(" -h       this help\n");
  printf(" -f usec  fast delay (%dus < fast < slow)\n", JM_FAST_LIMIT);
  printf(" -s usec  slow delay (fast < slow < %dus)\n", JM_SLOW_LIMIT);
  printf(" -X axis  select joystick axis for horizontal mouse movement\n");
  printf(" -Y axis  select joystick axis for vertical mouse movement\n");
  printf(" -l key   select the joystick button for left click\n");
  printf(" -r key   select the joystick button for right click\n");
  printf(" -m key   select the joystick button for middle click\n");
  printf(" -b key   select the handbrake button (stops drift when centered)\n");

  printf("\n");

  printf("joystick axes:\n");
  for(i = 0; abs_table[i].name; i++){
    printf(" %s", abs_table[i].name);
  }
  printf("\n\n");

  printf("joystick buttons/keys:\n");
  for(i = 1; button_table[i].name; i++){
    printf(" %s", button_table[i].name);
  }
  printf("\n");
}

static void device_clue(char *app, char *file, int code, char *module)
{
  char *msg;

  switch(code){
    case EACCES :
      fprintf(stderr, "%s: error: insufficient privileges to open device file %s\n", app, file);
      fprintf(stderr, "%s: suggest: running as root or changing device permissions/ownership\n", app);
      break;
    case ENOENT :
      fprintf(stderr, "%s: error: device file %s does not exist\n", app, file);
      if(strncmp("/dev/input", file, 10)){
        fprintf(stderr, "%s: suggest: using the appropriate file in /dev/input/\n", app);
      } else {
        fprintf(stderr, "%s: suggest: \"mknod %s c major minor\" as per README\n", app, file);
      }
      break;
    case ENODEV :
      fprintf(stderr, "%s: error: no device or driver backing file %s\n", app, file);
      fprintf(stderr, "%s: suggest: loading module with \"modprobe %s\" or plugging in device\n", app, module);
      break;
    default :
      fprintf(stderr, "%s: error: unable to open device file %s\n", app, file);
      msg = strerror(code);
      if(msg){
        fprintf(stderr, "%s: error: operating system reports \"%s\"\n", app, msg);
      } else {
        fprintf(stderr, "%s: error: operating system reports code %d\n", app, code);
      }
      break;
  }
}

static int make_rodent(char *app, int make)
{
  int fd, i;
  struct uinput_user_dev ui;
  unsigned int events[] = { EV_SYN, EV_KEY, EV_REL, EV_MAX };
  unsigned int mouse[] = { REL_X, REL_Y, REL_MAX};
  unsigned int keys[] = { BTN_LEFT, BTN_RIGHT, BTN_MIDDLE, KEY_MAX };
#if 0
  struct stat st;

  if(make){
    if(stat(device, &st) && (errno == ENODEV)){
      /* TODO ? use mknod to create the UINPUT device file */
    }
  }
#endif

  fd = open(JM_DEFAULT_UINPUT, O_RDWR | O_NONBLOCK);
  if(fd < 0){
    if(errno != ENOENT){
      device_clue(app, JM_DEFAULT_UINPUT, errno, "uinput");
      return -1;
    }
    fd = open(JM_FALLBACK_UINPUT, O_RDWR | O_NONBLOCK);
    if(fd < 0){
      device_clue(app, (errno == ENOENT) ? JM_DEFAULT_UINPUT : JM_FALLBACK_UINPUT, errno, "uinput");
      return -1;
    }
  }

  snprintf(ui.name, UINPUT_MAX_NAME_SIZE, "Marc's Joyevmouse (%d)", getpid());
  ui.name[UINPUT_MAX_NAME_SIZE - 1] = '\0';

  ui.id.bustype = 0;
  ui.id.vendor = 0;
  ui.id.product = 0;
  ui.id.version = 0;

  ui.ff_effects_max = 0;

  for(i = 0; i < ABS_MAX; i++){
    ui.absmax[i] = 0;
    ui.absmin[i] = 0;
    ui.absfuzz[i] = 0;
    ui.absflat[i] = 0;
  }
  
  for(i = 0; events[i] != EV_MAX; i++){
    if(ioctl(fd, UI_SET_EVBIT, events[i])){
      close(fd);
      return -1;
    }
  }

  for(i = 0; keys[i] != KEY_MAX; i++){
    if(ioctl(fd, UI_SET_KEYBIT, keys[i])){
      close(fd);
      return -1;
    }
  }

  for(i = 0; mouse[i] != REL_MAX; i++){
    if(ioctl(fd, UI_SET_RELBIT, mouse[i])){
      close(fd);
      return -1;
    }
  }

  if(write(fd, &ui, sizeof(struct uinput_user_dev)) != sizeof(struct uinput_user_dev)){
    close(fd);
    return -1;
  }

  if(ioctl(fd, UI_DEV_CREATE, 0)){
    close(fd);
    return -1;
  }

  return fd;
}

static int check_stick(char *app, int fd, int verbose)
{
  char *ptr;
  int version;
  int cap_abs, cap_key;
  char name[JM_NAME_SIZE];
  int i, j;
  unsigned long k;
  unsigned long mask[NBITS(EV_MAX)];

  if(ioctl(fd, EVIOCGVERSION, &version) < 0){
    fprintf(stderr, "%s: error: version ioctl failed: %s\n", app, strerror(errno));
    return -1;
  }
  if(verbose > 1){
    printf("%s: device version %d\n", app, version);
  }

  if(ioctl(fd, EVIOCGNAME(JM_NAME_SIZE), name) < 0){
    fprintf(stderr, "%s: error: name ioctl failed: %s\n", app, strerror(errno));
    return -1;
  }
  if(verbose){
    printf("%s: identified device: %s\n", app, name);
  }

  for(i = 0; i < NBITS(EV_MAX); i++){
    mask[i] = 0;
  }
  if(ioctl(fd, EVIOCGBIT(0, NBITS(EV_MAX)), mask) < 0){
    fprintf(stderr, "%s: error: unable to get device capabilities: %s\n", app, strerror(errno));
    return -1;
  }

  cap_key = 0;
  cap_abs = 0;
  if(verbose) {
    printf("%s: device capabilities:", app);
  }
  for(i = 0; i < NBITS(EV_MAX); i++){
    for(j = 0; j < (8 * sizeof(unsigned long)) ; j++){
      if((1UL << j) & mask[i]){
        k = (i * sizeof(unsigned long) * 8) + j;
        switch(k){
          case EV_KEY:
            cap_key = 1;
            break;
          case EV_ABS:
            cap_abs = 1;
            break;
        }
        if(verbose){
          ptr = code_to_name(cap_table, k);
          if(ptr){
            printf(" %s", ptr);
          } else {
            printf(" unknown(0x%lx)", k);
          }
        }
      }
    }
  }
  if(verbose){
    printf("\n");
  }

  /* tautology, but makes it explicit */
  return (cap_abs && cap_key) ? 1 : 0;
}

static int make_bin(int alpha, int beta)
{
  int result;

  result = (alpha - beta) / 3;
  if(result > 0){
    return result;
  }

  return 1;
}

/* lets hope the compiler optimises this function away */

static int make_move(int fd, int dim, int ctr, int pos, int bin, int alt, int min, int max)
{
  struct input_event vt;
  int range, stall, dir, more;

  vt.type = EV_REL;
  vt.code = dim;

  if(pos > ctr){
    dir = 1;
    range = pos - ctr;
    more = pos - alt;
  } else {
    dir = (-1);
    range = ctr - pos;
    more = alt - pos;
  }

  switch(range / bin){
    case 0 :
      vt.value = dir;
      stall = max; 
      /* stall = (140 + range) * 1000; */
      break;
    case 1 :
      vt.value = dir * ((range / 2) + 1);
      stall = (min + max + max) / 3;
      break;
    case 2 :
      vt.value = dir * range;
      stall = (min + max) / 2;
      break;
    default :
      vt.value = dir * more;
      stall = min;
      break;
  }

#ifdef DEBUG
  fprintf(stderr, "move: current=%d, center=%d, bin=%d, vt.value=%d, stall=%d\n", pos, ctr, bin, vt.value, stall);
#endif

  gettimeofday(&(vt.time), NULL);
  write(fd, &vt, sizeof(struct input_event));

  return stall;
}

int main(int argc, char **argv)
{
  int fd, ed;
  int left, right, middle, handbrake, vertical, horizontal;
  int i, j, c, stall, ok, result, value, run, syn;
  int tx, bx, mx, vx, lx, ux, dux, dlx, xw;
  int ty, by, my, vy, ly, uy, duy, dly, yw;
  fd_set fsr;
  struct timeval tv;
  struct input_event vt;
  int verbose, brittle;
  char *device, *app, *look;
  int fast, slow;

  brittle = 1;
  verbose = 1;
  device = NULL;
  app = argv[0];
  i = j = 1;

  slow = JM_SLOW_STALL;
  fast = JM_FAST_STALL;

  left = BTN_TRIGGER;
  right = BTN_THUMB;
  middle = BTN_TOP;
  handbrake = BTN_BASE;
  vertical = ABS_Y;
  horizontal = ABS_X;

  while (i < argc) {
    if (argv[i][0] == '-') {
      switch (argv[i][j]) {
        case 'c' :
          printf("(c) 2003,2005,2007,2008 Marc Welz\nLicensed under the conditions of the GNU General Public License\n");
          return EX_OK;
        case 'h' :
          usage(app);
          return EX_OK;
        case 'l': 
        case 'r': 
        case 'm': 
        case 'b': 
          c = argv[i][j];
          j++;
          if (argv[i][j] == '\0') {
            j = 0;
            i++;
          }
          if (i >= argc) {
            fprintf(stderr, "%s: error: option -%c needs a parameter\n", app, c);
            return EX_USAGE;
          }
          value = name_to_code(button_table, argv[i] + j);
          if(value == 0){
            value = atoi(argv[i] + j);
          }
          if(value == 0){
            fprintf(stderr, "%s: error: unknown joystick button: %s\n", app, argv[i] + j);
            return EX_USAGE;
          }
          switch(c){
            case 'l' : left = value; break;
            case 'r' : right = value; break;
            case 'm' : middle = value; break;
            case 'b' : handbrake = value; break;
          }
          i++;
          j = 1;
          break;
        case 'X': 
        case 'Y': 
          c = argv[i][j];
          j++;
          if (argv[i][j] == '\0') {
            j = 0;
            i++;
          }
          if (i >= argc) {
            fprintf(stderr, "%s: error: option -%c needs a parameter\n", app, c);
            return EX_USAGE;
          }
          value = name_to_code(abs_table, argv[i] + j);
          if(value == 0){
            value = atoi(argv[i] + j);
          }
          if(value == 0){
            fprintf(stderr, "%s: error: unknown joystick movement: %s\n", app, argv[i] + j);
            return EX_USAGE;
          }
          switch(c){
            case 'X' : horizontal = value; break;
            case 'Y' : vertical = value; break;
          }
          i++;
          j = 1;
          break;
        case 'f': 
        case 's': 
          c = argv[i][j];
          j++;
          if(argv[i][j] == '\0'){
            j = 0;
            i++;
          }
          if(i >= argc){
            fprintf(stderr, "%s: error: option -%c needs a parameter\n", app, c);
            return EX_USAGE;
          }
          value = atoi(argv[i] + j);
          if(value < JM_FAST_LIMIT){
            value = JM_FAST_LIMIT;
            fprintf(stderr, "%s: warning: limiting -%c delay to more than %dus\n", app, c, value);
          } else if(value > JM_SLOW_LIMIT){
            value = JM_SLOW_LIMIT;
            fprintf(stderr, "%s: warning: limiting -%c delay to less than %dus\n", app, c, value);
          }
          switch(c){
            case 'f' : fast = value; break;
            case 's' : slow = value; break;
          }
          i++;
          j = 1;
          break;
        case 'x' :
          brittle = brittle - 1;
          j++;
          break;
        case 'q' :
          verbose = 0;
          j++;
          break;
        case 'v' :
          verbose++;
          j++;
          break;
        case '-' :
          j++;
          break;
        case '\0':
          j = 1;
          i++;
          break;
        default:
          fprintf(stderr, "%s: error: unknown option -%c, use -h to see valid options\n", app, argv[i][j]);
          return EX_USAGE;
      }
    } else {
      if(device){
        fprintf(stderr, "%s: error: extra parameter %s, use -h for help\n", app, argv[i]);
        return EX_USAGE;
      } else {
        device = argv[i];
      }
      i++;
    }
  }

  if(device == NULL){
    fprintf(stderr, "%s: error: insufficient parameters, use -h for help\n", app);
    return EX_USAGE;
  }

  if(verbose){
    printf("%s: joystick to mouse converter (via evdev+uinput)\n", app);
  }

  ed = open(device, O_RDWR);
  if(ed < 0){
    device_clue(app, device, errno, "evdev");
    return EX_OSERR;
  }
  if(verbose > 1){
    printf("%s: opened physical device on %s\n", app, device);
  }

  result = check_stick(app, ed, verbose);
  if(result < 0){
    return EX_OSERR;
  }

  if(result == 0){
    fprintf(stderr, "%s: %s: no key or absolute capabilities, probably not a joystick\n", app, brittle ? "error" : "warning");
    if(brittle){
      return EX_USAGE;
    }
  }

  fd = make_rodent(app, 1);
  if(fd < 0){
    return EX_OSERR;
  }
  if(verbose > 1){
    printf("%s: created synthetic device via uinput\n", app);
  }

  stall = 0;
  run = 0;
  tv.tv_sec = 0;

  yw = xw = 1;
  vx = vy = 0;

  tx = ty = 0;
  bx = by = 0;
  syn = 1;

  lx = ux = ly = uy = 0;

  dux = duy = dlx = dly = 1;

  if(slow < fast){
    fprintf(stderr, "%s: warning: fast delay slower than slow, exchanging delays\n", app);
    value = slow;
    slow = fast;
    fast = value;
  }

  if(verbose){

    printf("%s: info: left click:   %s (%d)\n", app, code_to_name(button_table, left), left);
    printf("%s: info: right click:  %s (%d)\n", app, code_to_name(button_table, right), right);
    printf("%s: info: middle click: %s (%d)\n", app, code_to_name(button_table, middle), middle);
    printf("%s: info: handbrake:    %s (%d)\n", app, code_to_name(button_table, handbrake), handbrake);
    if(verbose > 1){

      printf("%s: info: horizontal:   %s (%d)\n", app, code_to_name(abs_table, horizontal), horizontal);
      printf("%s: info: vertical:     %s (%d)\n", app, code_to_name(abs_table, vertical), vertical);
      printf("%s: info: fast delay:   %dus\n", app, fast);
      printf("%s: info: slow delay:   %dus\n", app, slow);
    }

    printf("%s: info: please move joystick through full range\n", app);
    printf("%s: info: when finished return to central position\n", app); 
    printf("%s: info: and press handbrake (%s) to start\n", app, code_to_name(button_table, handbrake));
    printf("%s: notice: once started, the handbrake button should only be\n", app);
    printf("%s: notice: pressed when the pointer drifts in the neutral position\n", app);
  }

  while(1){
    FD_ZERO(&fsr);
    FD_SET(ed, &fsr);

    tv.tv_usec = stall;
    result = select(ed + 1, &fsr, NULL, NULL, stall ? &tv : NULL);

    if(result > 0){
      if(read(ed, &vt, sizeof(struct input_event)) == sizeof(struct input_event)){
        if(verbose > 3){
          fprintf(stderr, "%s: event: type=%d, code=%d, value=%d\n", app, vt.type, vt.code, vt.value);
        }

        ok = 0;
         
        switch(vt.type){
          case EV_SYN:
            if(syn == 0){
              ok = 1;
              syn = 1;
            }
            break;
          case EV_KEY :
            if(verbose > 1){
              look = code_to_name(button_table, vt.code);
              if(look){
                fprintf(stderr, "%s: button: %s", app, look);
              } else {
                fprintf(stderr, "%s: button: %d", app, vt.code);
              }
            }
            if(vt.code == left){
              if(verbose > 1){
                fprintf(stderr, "->left-click\n");
              }
              vt.code = BTN_LEFT;
              ok = 1;
              syn = 0;
            } else if(vt.code == right){
              if(verbose > 1){
                fprintf(stderr, "->right-click\n");
              }
              vt.code = BTN_RIGHT;
              ok = 1;
              syn = 0;
            } else if(vt.code == middle){
              if(verbose > 1){
                fprintf(stderr, "->middle-click\n");
              }
              vt.code = BTN_MIDDLE;
              ok = 1;
              syn = 0;
            } else if(vt.code == handbrake){
              if(verbose > 1){
                fprintf(stderr, "->handbrake\n");
              }
              if((yw == 0) && (xw == 0) && (vt.value == 1)){
                if(run == 0){
                  ly = uy = vy;
                  lx = ux = vx;
                  if(verbose){
                    fprintf(stderr, "%s: info: now ready to move mouse\n", app);
                  }
                  run = 1;
                }
                if(vx <= lx){
                  lx = vx;
                  dlx = make_bin(lx, bx);
                }
                if(vx >= ux){
                  ux = vx;
                  dux = make_bin(tx, ux);
                }
                if(vy <= ly){
                  ly = vy;
                  dly = make_bin(ly, by);
                }
                if(vy >= uy){
                  uy = vy;
                  duy = make_bin(ty, uy);
                }
                if(verbose > 1){
                  fprintf(stderr, "%s: center box: %d-%d, %d-%d\n", app, lx, ux, ly, uy);
                }
              }
            } else {
              if(verbose > 1){
                fprintf(stderr, "\n");
              }
            }
            break;
          case EV_ABS :
            if(vt.code == horizontal){
              vx = vt.value;
              if(vx > tx){
                tx = vx;
                mx = (tx - bx) / 2;
              }
              if(vx < bx){
                bx = vx;
                mx = (tx - bx) / 2;
              }
              if(xw){
                bx = tx = vx;
                xw = 0;
              }
            } else if(vt.code == vertical){
              vy = vt.value;
              if(vy > ty){
                ty = vy;
                my = (ty - by) / 2;
              }
              if(vy < by){
                by = vy;
                my = (ty - by) / 2;
              }
              if(yw){
                by = ty = vy;
                yw = 0;
              }
            }
            if(verbose > 1){
              fprintf(stderr, "%s: range: X=%d-%d, Y=%d-%d\n", app, bx, tx, by, ty);
            }

            break;
        }

        if(ok){
          if(verbose > 1){
            fprintf(stderr, "%s: sending event code=%d\n", app, vt.code);
          }
          gettimeofday(&(vt.time), NULL);
          write(fd, &vt, sizeof(struct input_event));
        } 
      }
    }

    if(run){

      stall = 0;

      if(vx < lx){
        stall = make_move(fd, REL_X, lx, vx, dlx, ux, fast, slow);
      }

      if(vx > ux){
        stall = make_move(fd, REL_X, ux, vx, dux, lx, fast, slow);
      }

      if(vy < ly){
        value = make_move(fd, REL_Y, ly, vy, dly, uy, fast, slow);
        if((stall <= 0) || (value < stall)){
          stall = value;
        }
      }

      if(vy > uy){
        value = make_move(fd, REL_Y, uy, vy, duy, ly, fast, slow);
        if((stall <= 0) || (value < stall)){
          stall = value;
        }
      }

#ifdef DEBUG
      fprintf(stderr, "stall: %d\n", stall);
#endif

      if(stall > 0){
        vt.value = 0;
        vt.code = 0;
        vt.type = EV_SYN;
        gettimeofday(&(vt.time), NULL);
        write(fd, &vt, sizeof(struct input_event));
        syn = 1;
      }
    }
  }

  return EX_OK;
}
